"""
The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.

 Author: wh0ami
License: MIT License <https://opensource.org/license/MIT>
Project: https://codeberg.org/wh0ami/hcloud-rdns-manager
"""

import argparse
import sys
from argparse import RawTextHelpFormatter
from pathlib import Path

import yaml
from loguru import logger

from hcloud_rdns_manager.deployment import deploy_changes
from hcloud_rdns_manager.export import generate_dump
from hcloud_rdns_manager.validation import validate_config


def main() -> None:
    """argparser, config loading, etc."""
    # read cli parameters
    parser = argparse.ArgumentParser(
        description="The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "--commit",
        action="store_true",
        help="Commit, that the changes should really be deployed. (only deploy action)",
    )
    parser.add_argument(
        "--servers",
        metavar="server-id",
        type=str,
        action="append",
        default=[],
        help="Only affect these server id(s). Can be given multiple times. By default, all servers are affected.",
    )
    parser.add_argument(
        "--projects",
        metavar="project-name",
        type=str,
        action="append",
        default=[],
        help="Only affect these projects. The project name passed here is NOT the name from the Web "
        "UI, it is the name that you defined in your local rdns zone file. Can be given multiple "
        "times. By default, all projects are affected.",
    )
    parser.add_argument(
        "action",
        type=str,
        metavar="action",
        choices=["check", "deploy", "dump"],
        help="Valid actions are:\n"
        "check\t- read the rdns zone file and validate it (--servers and --projects will be "
        "ignored, if passed)\n"
        "deploy\t- deploy the rdns entries from the rdns zone file in the Hetzner cloud, "
        "if --commit was passed. Otherwise, it will only perform a dry-run and display the "
        "changes.\n"
        "dump\t- obtain and dump the current config from the Hetzner cloud\n\n",
    )
    parser.add_argument("rdns_zone_file", type=str, help="rDNS zone file that should be processed")
    args = parser.parse_args(sys.argv[1:])

    # open and read config
    try:
        with Path(args.rdns_zone_file).open() as f:
            config = yaml.safe_load(f)
    except OSError:
        logger.error("Config not found or unreadable! Aborting...")
        sys.exit(1)
    except yaml.YAMLError:
        logger.error("Config contains invalid yaml! Aborting...")
        sys.exit(1)

    # validate the configuration (independent of action that should be executed)
    if validate_config(config):
        # check/execute the passed action
        if args.action == "deploy":
            deploy_changes(config=config, project_names=args.projects, server_ids=args.servers, commit=args.commit)
        elif args.action == "dump":
            generate_dump(config, project_names=args.projects, server_ids=args.servers)
        elif args.action == "check":
            pass
        else:
            logger.error("Invalid action!")
            sys.exit(1)
        sys.exit(0)
    else:
        sys.exit(1)



"""
The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.

 Author: wh0ami
License: MIT License <https://opensource.org/license/MIT>
Project: https://codeberg.org/wh0ami/hcloud-rdns-manager
"""

from datetime import UTC, datetime
from pathlib import Path

import yaml
from loguru import logger

from hcloud_rdns_manager.hcloudapi import HcloudApi
from hcloud_rdns_manager.validation import validate_config


def generate_dump(config: dict, project_names: list[str], server_ids: list[str]) -> bool:
    """
    Generate a dump of a currently provisioned configuration in the Hetzner Cloud.

    :param config: Configuration to use.
    :param project_names: Restrict action to the passed list of projects names.
    :param server_ids: Restrict action to the passed list of server IDs.
    :return: Whether the dump generation was successful or not.
    """
    # generate dump
    logger.info("Creating project dump...")

    # initialize empty export dump
    dump = {
        "hcloud": {
            "projects": {},
        },
    }

    # iterate over all projects
    for project in config["hcloud"]["projects"]:
        # check whether project should be dumped
        if len(project_names) == 0 or project in project_names:
            # fetch project config
            project_config = config["hcloud"]["projects"][project]

            # initialize api client
            api = HcloudApi(project_config["token"])

            # initialize project dump
            project_dump = {
                "token": str(project_config["token"]),
                "servers": {},
            }

            # iterate over list of servers, that should be dumped, dump then and append the dump to the project dump
            # filter from --servers is passed by serverList parameter
            for server in api.get_all_hcloud_server_ids(server_list=server_ids):
                project_dump["servers"][server] = {}
                project_dump["servers"][server]["rdns"] = api.get_hcloud_server_ptrs(server_id=server)

            # append project dump to export dump
            dump["hcloud"]["projects"][project] = project_dump

    logger.info("Dump created!")

    # validate generated dump
    valid = validate_config(config=dump)

    # write project dump into file, if its valid
    if valid:
        # generate name of dump file
        dumpfile = "hcloud_rdns_dump_" + datetime.now(tz=UTC).strftime("%Y-%m-%d_%H-%M-%S") + ".yml"

        # try to write dump as yaml into the dump file defined above
        logger.info(f"Writing dump to {dumpfile}...")
        try:
            with Path(dumpfile).open("w") as f:
                yaml.dump(dump, f)
        except OSError as ioe:
            logger.error(
                f"Could not write dump! Maybe your current working directory is not writeable? Aborting... [{ioe}]",
            )
            return False
        except yaml.YAMLError as exc:
            logger.error(f"Dump contains invalid yaml! Aborting... [{exc}]")
            return False

        # only on success
        logger.info(f"Dump written to {dumpfile} !")
        return True

    return False



"""
The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.

 Author: wh0ami
License: MIT License <https://opensource.org/license/MIT>
Project: https://codeberg.org/wh0ami/hcloud-rdns-manager
"""

import jsonschema
from loguru import logger


def validate_config(config: dict) -> bool:
    """
    Validate a configuration file by using jsonschema.

    :param config: The config, which should be validated.
    :return: Whether the validation was successful.
    """
    logger.info("Validating passed rdns zone config...")

    # define a schema for the configuration
    schema = {
        "type": "object",
        "additionalProperties": False,
        "required": ["hcloud"],
        "properties": {
            "hcloud": {
                "type": "object",
                "additionalProperties": False,
                "required": ["projects"],
                "properties": {
                    "projects": {
                        "type": "object",
                        "additionalProperties": False,
                        "minProperties": 1,
                        "patternProperties": {
                            "^[a-zA-Z0-9-]+$": {
                                "type": "object",
                                "additionalProperties": False,
                                "required": ["token", "servers"],
                                "properties": {
                                    "token": {
                                        "type": "string",
                                        "pattern": "^[a-zA-Z0-9]{64}$",
                                    },
                                    "servers": {
                                        "type": "object",
                                        "additionalProperties": False,
                                        "minProperties": 1,
                                        "patternProperties": {
                                            "^[0-9]+$": {
                                                "type": "object",
                                                "additionalProperties": False,
                                                "required": ["rdns"],
                                                "properties": {
                                                    "rdns": {
                                                        "type": "object",
                                                        "additionalProperties": False,
                                                        "patternProperties": {
                                                            "^([0-9\\.]+|[0-9a-fA-F:]+)$": {
                                                                "type": "string",
                                                            },
                                                        },
                                                    },
                                                },
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    }

    # validate the schema, quit on error
    try:
        jsonschema.validate(instance=config, schema=schema)
    except jsonschema.exceptions.ValidationError as error:
        logger.error("The rdns zone config is invalid: " + str(error))
        return False
    except Exception as error:  # noqa: BLE001
        logger.error("Error while validating the rdns zone config: " + str(error))
        return False

    # only on success
    logger.info("The passed rdns zone config is valid!")
    return True



"""
The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.

 Author: wh0ami
License: MIT License <https://opensource.org/license/MIT>
Project: https://codeberg.org/wh0ami/hcloud-rdns-manager
"""



"""
The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.

 Author: wh0ami
License: MIT License <https://opensource.org/license/MIT>
Project: https://codeberg.org/wh0ami/hcloud-rdns-manager
"""

from loguru import logger
from tabulate import tabulate

from hcloud_rdns_manager.hcloudapi import HcloudApi


def deploy_changes(config: dict, project_names: list, server_ids: list, commit: bool = False) -> None:  # noqa: C901
    """
    Deploy a passed configuration to the Hetzner Cloud.

    :param config: The configuration to use.
    :param project_names: Restrict the action to the passed list of project names.
    :param server_ids: Restrict the action to the passed list of server IDs.
    :param commit: Whether the changes should be deployed or just showed.
    """
    # iterate over all projects
    for project in config["hcloud"]["projects"]:
        # check whether project should be updated
        if len(project_names) == 0 or project in project_names:
            # fetch project config
            project_config = config["hcloud"]["projects"][project]

            # initialize api client
            api = HcloudApi(project_config["token"])

            # fetch current list of server ids (--servers will be applied by passing servers parameter)
            id_list = api.get_all_hcloud_server_ids(server_list=server_ids)

            # iterate over all servers that were fetched above
            for server in id_list:
                # just take those servers, which are defined in rdns zone config
                if server in project_config["servers"]:
                    # fetch current rdns configuration from hcloud
                    current_rdns = api.get_hcloud_server_ptrs(server_id=server)

                    # initialize empty changeset
                    changeset = []

                    # iterate over all entries, that should exist (as defined in config)
                    for ip in project_config["servers"][server]["rdns"]:
                        # ptr that doesn't exist in the current_rdns list needs to be created
                        if ip not in current_rdns:
                            changeset.append(["CREATE", ip, None, project_config["servers"][server]["rdns"][ip]])
                        # ptr that exist (logical result of the condition above) but do not have the same value needs to
                        # be updated
                        elif current_rdns[ip] != project_config["servers"][server]["rdns"][ip]:
                            changeset.append(
                                ["UPDATE", ip, current_rdns[ip], project_config["servers"][server]["rdns"][ip]],
                            )

                    # iterate over all entries, that currently exist
                    changeset.extend(
                        ["DELETE", ip, current_rdns[ip], None]
                        for ip in current_rdns
                        if ip not in project_config["servers"][server]["rdns"]
                    )

                    # print changeset as table
                    logger.info(
                        f"--> Server {server} ({api.get_hcloud_server_nickname_by_id(server_id=server)}) "
                        f"from project {project}",
                    )
                    if len(changeset) > 0:
                        logger.info(
                            "Changeset:\n"
                            + tabulate(changeset, headers=["Action", "IP", "Current PTR", "New PTR"], tablefmt="psql"),
                        )
                        if commit:
                            logger.info("Applying changeset...")
                            for change in changeset:
                                api.set_hcloud_server_ptr(server_id=server, ip=change[1], ptr=change[3])
                            logger.info("Changes applied!")
                    else:
                        logger.warning("Nothing to do!")



"""
The hcloud-rdns-manager - the smart way to manage your rDNS records in the Hetzner cloud.

 Author: wh0ami
License: MIT License <https://opensource.org/license/MIT>
Project: https://codeberg.org/wh0ami/hcloud-rdns-manager
"""

import hcloud.servers.client
from hcloud import Client


class HcloudApi:
    """Wrapper around of the Hetzner Cloud library."""

    def __init__(self, token: str) -> None:
        """
        Initialize api connection in ctor and store it as a private attribute.

        :param token: The API token to authenticate at the Hetzner Cloud API.
        """
        self.__client = Client(token=token)

    def get_all_hcloud_server_ids(self, server_list: list | None = None, cast_to_string: bool = True) -> list:
        """
        Get a list of all server ids from the api.

        :param server_list: Restrict action to this list of servers.
        :param cast_to_string: Whether the method should be explicitly cast the server IDs to strings.
        :return: List of server IDs.
        """
        if server_list is None:
            server_list = []
        id_list = []
        for server in self.__client.servers.get_all():
            server_id = str(server.id) if cast_to_string else server.id
            id_list.append(server_id) if len(server_list) == 0 or server_id in server_list else None
        return id_list

    def get_hcloud_server_ptrs(self, server_id: str) -> dict:
        """
        Fetch a dict with all ptr records of a server, which will be identified by its id.

        :param server_id: ID of the server, of which the PTRs should be returned.
        :return: Dict with PTRs.
        """
        server = self.__client.servers.get_by_id(id=server_id)
        ptrs = {server.public_net.ipv4.ip: server.public_net.ipv4.dns_ptr}
        for ptr in server.public_net.ipv6.dns_ptr:
            ptrs[ptr["ip"]] = ptr["dns_ptr"]
        return ptrs

    def set_hcloud_server_ptr(self, server_id: str, ip: str, ptr: str) -> hcloud.servers.client.BoundAction:
        """
        Update the ptr of a server.

        :param server_id: ID of the server, of which the PTR should be set.
        :param ip: IP address, for which the PTR should be created.
        :param ptr: The actual PTR (DNS name).
        :return: BoundAction from the hcloud library.
        """
        return self.__client.servers.get_by_id(id=server_id).change_dns_ptr(ip=ip, dns_ptr=ptr)

    def get_hcloud_server_nickname_by_id(self, server_id: str) -> str:
        """
        Fetch the custom nickname of a server from the api.

        :param server_id: IF of the server, of which the nickname should be returned.
        :return: The nickname of the server.
        """
        return self.__client.servers.get_by_id(id=server_id).name
